
-- process input from formspec created in get_fs_edit_option_dialog(..)
yl_speak_up.input_edit_option_dialog = function(player, formname, fields)
	if formname ~= "yl_speak_up:edit_option_dialog" then
		return
	end
        local pname = player:get_player_name()

	-- Is the player working on this particular npc?
	local edit_mode = (yl_speak_up.edit_mode[pname] == yl_speak_up.speak_to[pname].n_id)
	if(not(edit_mode)) then
		return
	end
	local n_id = yl_speak_up.speak_to[pname].n_id
	local d_id = yl_speak_up.speak_to[pname].d_id
	local o_id = yl_speak_up.speak_to[pname].o_id

	local dialog = yl_speak_up.speak_to[pname].dialog
	local n_dialog = dialog.n_dialogs[d_id]
	local d_option = n_dialog.d_options[o_id]

	if(fields.switch_tab and fields.switch_tab == "2") then
		yl_speak_up.show_fs(player, "edit_option_dialog",
			{n_id = n_id, d_id = d_id, o_id = o_id,
			 caller="show_if_action_failed"})
		return
	elseif(fields.switch_tab and fields.switch_tab == "1") then
		yl_speak_up.show_fs(player, "edit_option_dialog",
			{n_id = n_id, d_id = d_id, o_id = o_id,
			 caller="show_if_action_succeeded"})
		return
	end

	-- this menu is specific to an option for a dialog; if no dialog is selected, we really
	-- can't know what to do
	if(not(o_id) and d_id) then
		yl_speak_up.show_fs(player, "talk", {n_id = n_id, d_id = d_id})
	elseif(not(d_id)) then
		return
	end

	-- backwards compatibility to when this was a hidden field
	fields.o_id = o_id
	-- handles changes to o_text_when_prerequisites_met, target dialog, adding of a new dialog
	local result = yl_speak_up.edit_mode_apply_changes(pname, fields)
	-- if a new option was added or the target dialog of this one changed, display the right new option
	if(result and result["show_next_option"] and n_dialog.d_options[result["show_next_option"]]) then
		yl_speak_up.show_fs(player, "edit_option_dialog",
			{n_id = n_id, d_id = d_id, o_id = result["show_next_option"],
			 caller="show_next_option"})
		return
	end

	if(fields.save_option) then
		yl_speak_up.show_fs(player, "edit_option_dialog",
			{n_id = n_id, d_id = d_id, o_id = o_id, caller="save_option"})
		return
	end

	-- want to edit the text that is shown when switching to the next dialog?
	if(fields.button_edit_action_failed_dialog) then
		-- the target effect is the (failed) action
		local target_action = {}
		local actions = d_option.actions
		if(actions) then
			for a_id, a in pairs(actions) do
				if(a and a.a_id) then
					target_action = a
				end
			end
		end
		-- remember what we're working at
		yl_speak_up.speak_to[pname].edit_alternate_text_for = target_action
		yl_speak_up.show_fs(player, "msg", {
			input_to = formname,
			formspec = yl_speak_up.get_fs_edit_dialog_modification(
					dialog, target_action.a_on_failure, target_action.alternate_text,
					"if the action \""..tostring(target_action.a_id)..
					"\" of option \""..tostring(o_id)..
					"\" of dialog \""..tostring(d_id)..
					"\" failed because the player did something wrong")
			})
		return

	elseif(fields.button_edit_action_success_dialog) then
		-- the target effect is the "dialog" effect
		local target_effect = {}
		local results = d_option.o_results
		if(results) then
			for r_id, r in pairs(results) do
				if(r and r.r_type and r.r_type == "dialog") then
					target_effect = r
				end
			end
		end
		-- remember what we're working at
		yl_speak_up.speak_to[pname].edit_alternate_text_for = target_effect
		yl_speak_up.show_fs(player, "msg", {
			input_to = formname,
			formspec = yl_speak_up.get_fs_edit_dialog_modification(
					dialog, target_effect.r_value, target_effect.alternate_text,
					"if the action "..
					"of option \""..tostring(o_id)..
					"\" of dialog \""..tostring(d_id)..
					"\" was successful - or if there was no action")
			})
		return

	-- save the alternate action/effect text for the next dialog
	elseif(fields.save_dialog_modification) then
		local target_element = yl_speak_up.speak_to[pname].edit_alternate_text_for
		local old_text = target_element.alternate_text
		if(target_element and fields.d_text_new and fields.d_text_new ~= "$TEXT$"
		  and fields.d_text_new ~= target_element.alternate_text) then
			target_element.alternate_text = fields.d_text_new
			if(target_element and target_element.r_id) then
				-- record the change
				table.insert(yl_speak_up.npc_was_changed[ n_id ],
					"Dialog "..d_id..": The text displayed for dialog "..
					tostring(target_element.r_value).." when selecting option "..
					tostring( o_id ).." in dialog "..tostring( d_id )..
					" was changed from "..
					"["..tostring(old_text).."] to ["..tostring(fields.d_text_new).."].")
			elseif(target_element and target_element.a_id) then
				-- record the change
				table.insert(yl_speak_up.npc_was_changed[ n_id ],
					"Dialog "..d_id..": The text displayed for dialog "..
					tostring(target_element.a_on_failure).." when the action "..
					tostring(target_element.a_id).." of option "..
					tostring( o_id ).." in dialog "..tostring( d_id )..
					" failed, was changed from "..
					"["..tostring(old_text).."] to ["..tostring(fields.d_text_new).."].")
			end
		end

		yl_speak_up.show_fs(player, "edit_option_dialog",
			{n_id = n_id, d_id = d_id, o_id = o_id, caller="back-from_edit_dialog_modifications"})
		return

	elseif(fields.back_from_edit_dialog_modification) then
		-- no longer working on an alternate text
		yl_speak_up.speak_to[pname].edit_alternate_text_for = nil
		yl_speak_up.show_fs(player, "edit_option_dialog",
			{n_id = n_id, d_id = d_id, o_id = o_id, caller="back_from_edit_dialog_modifications"})
		return
	end

	-- back to the main dialog window?
	-- (this also happens when the last option was deleted)
	if(fields.show_current_dialog or fields.quit or fields.button_exit or not(d_option) or fields.del_option) then
		yl_speak_up.show_fs(player, "talk", {n_id = n_id, d_id = d_id})
		return
	end

	-- the player wants to see the previous option/answer
	if(fields.edit_option_prev) then
		-- sort all options by o_sort
		local sorted_list = yl_speak_up.get_sorted_options(n_dialog.d_options, "o_sort")
		local o_found = o_id
		for i, o in ipairs(sorted_list) do
			if(o == o_id and sorted_list[ i-1]) then
				o_found = sorted_list[ i-1 ]
			end
		end
		-- show that dialog; fallback: show the same (o_id) again
		yl_speak_up.show_fs(player, "edit_option_dialog",
			{n_id = n_id, d_id = d_id, o_id = o_found, caller="prev option"})
		return

	-- the player wants to see the next option/answer
	elseif(fields.edit_option_next) then
		-- sort all options by o_sort
		local sorted_list = yl_speak_up.get_sorted_options(n_dialog.d_options, "o_sort")
		local o_found = o_id
		for i, o in ipairs(sorted_list) do
			if(o == o_id and sorted_list[ i+1 ]) then
				o_found = sorted_list[ i+1 ]
			end
		end
		-- show that dialog; fallback: show the same (o_id) again
		yl_speak_up.show_fs(player, "edit_option_dialog",
			{n_id = n_id, d_id = d_id, o_id = o_found, caller="next option"})
		return

	-- the player clicked on a precondition
	elseif(fields.table_of_preconditions) then
		yl_speak_up.show_fs(player, "edit_preconditions", fields.table_of_preconditions)
		return

	-- the player clicked on an action
	elseif(fields.table_of_actions) then
		yl_speak_up.show_fs(player, "edit_actions", fields.table_of_actions)
		return

	-- the player clicked on an effect
	elseif(fields.table_of_effects) then
		yl_speak_up.show_fs(player, "edit_effects", fields.table_of_effects)
		return
	end

	-- if ESC is pressed or anything else unpredicted happens: go back to the main dialog edit window
	-- reason: don't loose any unsaved changes to the dialog
	yl_speak_up.show_fs(player, "talk", {n_id = n_id, d_id = d_id})
end


-- edit options (not via staff but via the "I am your owner" dialog)
yl_speak_up.get_fs_edit_option_dialog = function(player, n_id, d_id, o_id, caller)
	-- n_id, d_id and o_id have already been checked when this function is called
	local pname = player:get_player_name()
	local dialog = yl_speak_up.speak_to[pname].dialog
	local n_dialog = dialog.n_dialogs[d_id]
	local d_option = n_dialog.d_options[o_id]

	-- currently no trade running (we're editing options)
	yl_speak_up.trade[pname] = nil
	yl_speak_up.speak_to[pname].trade_id = nil

	if(not(d_option)) then
		return "size[6,2]"..
			"label[0.2,0.5;Ups! Option "..minetest.formspec_escape(tostring(o_id))..
				" does not exist.]"..
			"button_exit[2,1.5;1,0.9;exit;Exit]"
	end
	-- remember which option we are working at (better than a hidden field)
	yl_speak_up.speak_to[pname].o_id = o_id
	-- are there any preconditions?
	local list_of_preconditions = ""
	local prereq = d_option.o_prerequisites
	local count_prereq = 0
	if(prereq) then
		local sorted_key_list = yl_speak_up.sort_keys(prereq)
		for i, k in ipairs(sorted_key_list) do
			local v = prereq[ k ]
			list_of_preconditions = list_of_preconditions..
				minetest.formspec_escape(v.p_id)..",#FFFF00,"..
				minetest.formspec_escape(v.p_type)..","..
				minetest.formspec_escape(
					yl_speak_up.show_precondition(v))..","
			count_prereq = count_prereq + 1
		end
	end
	if(count_prereq < yl_speak_up.max_prerequirements) then
		list_of_preconditions = list_of_preconditions..",#00FF00,add,Add a new pre(C)ondition"
	else
		list_of_preconditions = list_of_preconditions..",#AAAAAA,-,"..
			"Maximum amount of pre(C)onditions per option reached!"
	end

	-- build action list the same way as list of preconditions and effects
	local list_of_actions = ""
	local actions = d_option.actions
	local count_actions = 0
	local action_data = nil
	if(actions) then
		local sorted_key_list = yl_speak_up.sort_keys(actions)
		for i, k in ipairs(sorted_key_list) do
			local v = actions[ k ]
			list_of_actions = list_of_actions..
				minetest.formspec_escape(v.a_id)..",#FFFF00,"..
				minetest.formspec_escape(v.a_type)..","..
				minetest.formspec_escape(
					yl_speak_up.show_action(v))..","
			count_actions = count_actions + 1
			action_data = v
		end
	end
	if(count_actions < yl_speak_up.max_actions) then
		list_of_actions = list_of_actions..",#00FF00,add,Add a new (A)ction"
	else
		list_of_actions = list_of_actions..",#AAAAAA,-,"..
			"Maximum amount of (A)ctions per option reached!"
	end

	-- find the right target dialog for this option (if it exists)
	local target_dialog = nil
	-- which effect holds the information about the target dialog?
	-- set this to a fallback for yl_speak_up.show_colored_dialog_text
	local target_effect = {r_id = "-?-", r_type = "dialog"}
	-- and build the list of effects
	local list_of_effects = ""
	local results = d_option.o_results
	local count_effects = 0
	-- create a new dialog type option if needed
	if(not(results) or not(next(results))) then
		target_dialog = yl_speak_up.prepare_new_dialog_for_option(
			dialog, pname, n_id, d_id, o_id,
			yl_speak_up.text_new_dialog_id,
			results)
		-- make sure we are up to date (a new option was inserted)
		results = d_option.o_results
	end
	if(results) then
		local sorted_key_list = yl_speak_up.sort_keys(results)
		for i, k in ipairs(sorted_key_list) do
			local v = results[ k ]
			if v.r_type == "dialog" and dialog.n_dialogs[v.r_value] ~= nil then
				list_of_effects = list_of_effects..
					minetest.formspec_escape(v.r_id)..",#999999,"..
					minetest.formspec_escape(v.r_type)..","..
					minetest.formspec_escape(
						yl_speak_up.show_effect(v))..","
				-- there may be more than one in the data structure
				target_dialog = v.r_value
				target_effect = v
			elseif v.r_type ~= "dialog" then
				list_of_effects = list_of_effects..
					minetest.formspec_escape(v.r_id)..",#FFFF00,"..
					minetest.formspec_escape(v.r_type)..","..
					minetest.formspec_escape(
						yl_speak_up.show_effect(v))..","
			end
			count_effects = count_effects + 1
		end
	end
	if(count_effects < yl_speak_up.max_result_effects) then
		list_of_effects = list_of_effects..",#00FF00,add,Add a new (Ef)fect"
	else
		list_of_effects = list_of_effects..",#AAAAAA,-,"..
			"Maximum amount of allowed (Ef)fects per option reached!"
	end

	-- if no target dialog has been selected: default is to go to the dialog with d_sort 0
	if(not(target_dialog) or target_dialog == "" or not(dialog.n_dialogs[target_dialog])) then
		for d, v in pairs(dialog.n_dialogs) do
			if(v.d_sort and tonumber(v.d_sort) == 0) then
				target_dialog = d
			end
		end
	end
	-- build the list of available dialogs for the dropdown list(s)
	local dialog_list = yl_speak_up.text_new_dialog_id
	local dialog_selected = "1"
	-- if there are dialogs defined
	if(dialog and dialog.n_dialogs) then
		-- the first entry will be "New dialog"
		local n = 1
		for k, v in pairs(dialog.n_dialogs) do
			dialog_list = dialog_list .. "," .. minetest.formspec_escape(v.d_id)
			-- which one is the current dialog?
			n = n + 1
			if(v.d_id == target_dialog) then
				dialog_selected = tostring(n)
			end
		end
	end
	if(not(target_dialog)) then
		target_dialog = "- none -"
	end

	-- offer the correct preselection for hidden/grey/show text
	local alternate_answer_option = "3"
	if(d_option.o_hide_when_prerequisites_not_met == "true") then
		alternate_answer_option = "1"
	elseif(d_option.o_grey_when_prerequisites_not_met == "true") then
		alternate_answer_option = "2"
	end

	-- can the button "prev(ious)" be shown?
	local button_prev = ""
	-- can the button "next" be shown?
	local button_next = ""
	-- sort all options by o_sort
	local sorted_list = yl_speak_up.get_sorted_options(n_dialog.d_options, "o_sort")
	local o_found = o_id
	local anz_options = 0
	for i, o in ipairs(sorted_list) do
		-- the buttons are inside a container; thus, Y is 0.0
		if(o == o_id and sorted_list[ i-1 ]) then
			button_prev = ""..
				"button[7.9,0.0;2.0,0.9;edit_option_prev;Prev]"..
				"tooltip[edit_option_prev;Go to previous option/answer "..
				"(according to o_sort).]"
		end
		if(o == o_id and  sorted_list[ i+1 ]) then
			button_next = ""..
				"button[12.5,0.0;2.0,0.9;edit_option_next;Next]"..
				"tooltip[edit_option_next;Go to next option/answer "..
				"(according to o_sort).]"
		end
		anz_options = anz_options + 1
	end

	-- less than yl_speak_up.max_number_of_options_per_dialog options?
	local button_add = ""..
		-- the buttons are inside a container; thus, Y is 0.0
		"button[2.4,0.0;2.0,0.9;add_option;Add]"..
		"tooltip[add_option;Add a new option/answer to this dialog.]"
	if(anz_options >= yl_speak_up.max_number_of_options_per_dialog) then
		button_add = ""
	end

	-- make all following coordinates relative
	local action_text = "container[0.2,13.0]"..
		"box[0.25,0.0;21.0,6.7;#555555]"
	-- show what happens if the action fails
	if(caller == "show_if_action_failed") then
		action_text = action_text..
			-- allow to switch between successful and failed actions
			"tabheader[0.2,0.0;switch_tab;"..
				"If the action was successful:,"..
				"If the action failed:;2;true;true]"..
			"label[0.4,0.6;"..
			"If the player *failed* to complete the above action correctly,]"
		if(action_data and action_data.a_on_failure
		  and dialog.n_dialogs and dialog.n_dialogs[ action_data.a_on_failure]) then
			action_text = action_text..
			-- ..and what the NPC will reply to that answer
				"tooltip[1.2,3.9;19.6,2.5;This is what the NPC will say next when "..
					"the player has failed to complete the action.]"..

				"container[0.0,3.2]"..
				"label[0.4,0.4;..the NPC will react to this failed action with the "..
					"following dialog \""..tostring(action_data.a_on_failure)..
					"\""..
				yl_speak_up.show_colored_dialog_text(
					dialog,
					action_data,
					action_data.a_on_failure,
					"1.2,0.7;19.6,2.5;d_text_next",
					"with the *modified* text",
					":]",
					"button_edit_action_failed_dialog")..
				"container_end[]"
		else
			action_text = action_text..
				"label[0.4,3.6;..go back to the initial dialog.]"
		end
	-- show what happens if the action was successful
	else
		-- no action defined
		if(count_actions == 0) then
			-- do not show tabheader
			action_text = action_text..
				"label[0.4,0.6;"..
				"There is no (A)ction defined. Directly apply the following (Ef)fects:]"
		else
			action_text = action_text..
				-- allow to switch between successful and failed actions
				"tabheader[0.2,0.0;switch_tab;"..
					"If the action was successful:,"..
					"If the action failed:;1;true;true]"..
				"label[0.4,0.6;"..
					"If the player completed the above action successfully, "..
					"apply the following (Ef)fects:]"
		end
		action_text = action_text..
			-- list of effects
			"tablecolumns[text;color,span=1;text;text]"..
			"table[1.2,0.9;19.6,2.0;table_of_effects;"..
				list_of_effects..";0]"..
			"tooltip[1.2,0.9;19.6,2.0;"..
				"*All* (Ef)fects are executed after the action (if there is\n"..
				"one defined in this option) has been completed successfully\n"..
				"by the player. If there is no action defined, then the\n"..
				"(Ef)fects will always be executed when this option here is\n"..
				"selected.\n"..
				"Please click on an (Ef)fect in order to edit or delete it!]"..

			"container[0.0,3.2]"..
			-- allow to change the target dialog via a dropdown menu
			"dropdown[10.2,0.0;3.0,0.7;d_id_"..minetest.formspec_escape(o_id)..";"..
				dialog_list..";"..dialog_selected..",]"..
			"tooltip[10.2,0.0;3.0,0.7;Select the target dialog with which the NPC shall react "..
				"to this answer. Currently, dialog \""..
				minetest.formspec_escape(target_dialog)..
				"\" is beeing displayed.;#FFFFFF;#000000]"..
			-- ..and what the NPC will reply to that answer
			"tooltip[1.2,0.7;19.6,2.5;This is what the NPC will say next when the player has "..
				"selected this answer here.]"..
			"label[0.4,0.4;The NPC will react to this answer with dialog:]"..
			yl_speak_up.show_colored_dialog_text(
				dialog,
				-- this is either the "dialog" effect or an empty fallback
				target_effect,
				-- this is the text the NPC will say in reaction to this answer
				target_dialog,
				"1.2,0.7;19.6,2.5;d_text",
				"label[13.5,0.4;with the following *modified* text:]",
				"",
				"button_edit_action_success_dialog")..
				"container_end[]"
	end
	action_text = action_text.."container_end[]"

	-- build up the formspec
	local formspec = ""..
		"formspec_version[3]"..
		"size[22,21]"..
		"bgcolor[#00000000;false]"..
		-- button back to the current dialog (of which this is an option)
		"button[16.4,0.2;5.0,0.9;show_current_dialog;Back to dialog "..
			minetest.formspec_escape(d_id).."]"..
		"tooltip[show_current_dialog;Go back to dialog "..
			minetest.formspec_escape(d_id).." and continue editing that dialog.]"..
		-- tell the player what this formspec is about
		"label[6.5,0.4;You are editing dialog option \""..tostring(o_id).."\":]"..

		-- the text the NPC says
		"container[0.0,0.9]"..
		"label[0.2,0.0;NPC says "..
			minetest.formspec_escape("[dialog \""..tostring(d_id).."\"]:").."]"..
		yl_speak_up.show_colored_dialog_text(
			dialog,
			{r_id = "", r_type = "dialog"},
			d_id,
			"1.2,0.3;20.2,2.5;d_text",
			"", -- no modifications possible at this step
			"",
			"").. -- no edit button here as this text cannot be changed here
		"tooltip[1.2,0.3;20.2,3.0;This is what the NPC says to the player.]"..
		"container_end[]"..

		-- list the preconditions
		"container[0.0,4.4]"..
		"label[0.2,0.0;If all of the following pre(C)onditions are fulfilled:]"..
		"tablecolumns[text;color,span=1;text;text]"..
		"table[1.2,0.3;20.2,2.0;table_of_preconditions;"..
			list_of_preconditions..";0]"..
		"tooltip[1.2,0.3;20.2,2.0;"..
			"*All* pre(C)onditions need to be true in order\n"..
			"for the option to be offered to the player.\n"..
			"Please click on a pre(C)ondition in order\n"..
			"to edit or delete it!]"..
		"container_end[]"..

		-- answer of the player (the actual option)
		"container[0.0,7.3]"..
		"label[0.2,0.0;..the player may answer with this text"..
			minetest.formspec_escape(" [dialog option \""..tostring(o_id).."\"]:").."]"..
		"label[1.2,0.8;A:]"..
		"field[1.7,0.3;19.6,0.9;text_option_"..minetest.formspec_escape(o_id)..";;"..
			minetest.formspec_escape(d_option.o_text_when_prerequisites_met).."]"..
		"tooltip[option_text_met;This is the answer the player may choose if the "..
			"preconditions are all fulfilled.]"..
		-- dropdown for selecting weather to show the alternate answer or not
		"label[0.2,1.7;..but if at least one pre(C)ondition is not fulfilled, then...]"..
		"dropdown[12.0,1.3;9.3,0.7;hide_or_grey_or_alternate_answer;"..
			"..hide this answer.,"..
			"..grey out the following answer:,"..
			"..display the following alternate answer:;"..
			alternate_answer_option..";]"..
		-- alternate answer
		"label[1.2,2.5;A:]"..
		"field[1.7,2.0;19.6,0.9;option_text_not_met;;"..
			minetest.formspec_escape(d_option.o_text_when_prerequisites_not_met).."]"..
		"tooltip[option_text_not_met;This is the answer the player may choose if the "..
			"preconditions are NOT all fulfilled.]"..
		"container_end[]"..

		-- list of (A)ctions (there can only be one per option; i.e. a trade)
		"container[0.0,11.0]"..
		"label[0.2,0.0;When this answer has been selected, start the following (A)ction:]"..
		"tablecolumns[text;color,span=1;text;text]"..
		"table[1.2,0.3;20.2,0.7;table_of_actions;"..
			list_of_actions..";0]"..
		"container_end[]"..

		-- list effects and target dialog for successful - and target dialog for unsuccessful
		-- actions (including a toggle button)
		action_text..

		-- container for the buttons/footer
		"container[0.0,19.9]"..
		-- button: delete
		"button[0.2,0.0;2.0,0.9;del_option;Delete]"..
		"tooltip[del_option;Delete this option/answer.]"..
		-- button: add new
		button_add..
		-- button: save
		"button[4.6,0.0;2.0,0.9;save_option;Save]"..
		"tooltip[save_option;Save what you canged (or discard it).]"..
		-- button: prev/next
		button_prev..
		button_next..
		-- button: go back to dialog (repeated from top of the page)
		"button[15.8,0.0;5.0,0.9;show_current_dialog;Back to dialog "..
			minetest.formspec_escape(d_id).."]"..
		"tooltip[show_current_dialog;Go back to dialog "..
			minetest.formspec_escape(d_id).." and continue editing that dialog.]"..
		-- allow to enter o_sort
		"label[10.1,0.5;Sort:]"..
		"field[11.1,0.0;1.0,0.9;edit_option_o_sort;;"..
			minetest.formspec_escape(d_option.o_sort).."]"..
		"tooltip[edit_option_o_sort;o_sort: The lower the number, the higher up in the "..
			"list this option goes\nNegative values are ignored;#FFFFFF;#000000]"..
		"container_end[]"
	return formspec
end
